#!/bin/bash

echo "LMCache code quality tools (lining, formatting, spelling and static checks) are now managed by pre-commit hooks."
echo "Please run 'pip install -r requirements/lint.txt', followed by"
echo "'pre-commit install' to install the pre-commit hooks."
echo "Then code quality checks will run automatically before each commit."
echo "Check out https://pre-commit.com/ for more details."